#include <graph/server/gui/chart/chart.h>
#include <graph/server/gui/mainwindow.h>
#include <graph/server/gui/specdialog/specdialog.h>
#include <graph/server/plugin.h>
#include <graph/server/spec.h>
#include <graph/server/ui_mainwindow.h>
#include <qfiledialog.h>
#include <QFileDialog>
#include <QMessageBox>
#include <iostream>
#include <stdexcept>

#define MIN(a, b) (a < b ? a : b)
#define MAX(a, b) (a > b ? a : b)

namespace Graph::GUI {

MainWindow* MainWindow::instance = nullptr;

MainWindow* MainWindow::getInstance() {
  if (!instance)
    instance = new MainWindow();

  return instance;
}

void MainWindow::close() {
  SpecDialog::close();
  if (instance)
    delete instance;
}

void addData(unsigned int chartIndex, const char* seriesName, const point_t* points, int count) {
  MainWindow::getInstance()->addData(chartIndex, seriesName, points, count);
}

void MainWindow::addData(unsigned int chartIndex, const char* seriesName, const point_t* points, int count) {
  if (chartIndex < 0 || chartIndex >= charts.size())
    return;

  QString name(seriesName);

  ChartWidget* chart = charts[chartIndex];

  double minX = std::numeric_limits<double>::max(), maxX = std::numeric_limits<double>::min();
  double minY = std::numeric_limits<double>::max(), maxY = std::numeric_limits<double>::min();

  for (QAbstractSeries* series : chart->getSeries()) {
    if (series->name() == name) {
      for (int i = 0; i < count; i++) {
        *((QLineSeries*)series) << QPointF(points[i].x, points[i].y);
      }
    }

    for (QPointF point : ((QLineSeries*)series)->points()) {
      minX = MIN(minX, point.x());
      maxX = MAX(maxX, point.x());
      minY = MIN(minY, point.y());
      maxY = MAX(maxY, point.y());
    }
  }

  if (chart->getSpec().axis.x.auto_limit)
    chart->getX()->setRange(minX, maxX);
  if (chart->getSpec().axis.y.auto_limit)
    chart->getY()->setRange(minY, maxY);

  chart->getView()->update();
}

MainWindow::MainWindow(QWidget* parent) : QMainWindow(parent), ui(new Ui::MainWindow) {
  ui->setupUi(this);

  // plugins
  pluginModel = new PluginModel(this);
  ui->pluginsListView->setModel(pluginModel);

  // unload plugin
  connect(ui->pluginsListView, &QListView::doubleClicked, [&](const QModelIndex& index) {
    QMessageBox::StandardButton result = QMessageBox::question(this, "Unload plugin?", "Are you sure?", QMessageBox::Yes | QMessageBox::Cancel);

    if (result == QMessageBox::Yes) {
      pluginModel->removeRow(index.row());
      pluginModel->remove(index.row());
      ui->pluginsListView->doItemsLayout();
    }
  });

  fileDialog = new QFileDialog(this, "Load plugin", "./plugins", "Plugin (*.so *.dll *.dlsym)");
}

MainWindow::~MainWindow() {
  for (ChartWidget* chart : charts) {
    delete chart;
  }

  delete pluginModel;
  delete fileDialog;
}

void MainWindow::on_action_Add_triggered() {
  SpecDialog* dialog = SpecDialog::getInstance();
  if (dialog->exec() == SpecDialog::Rejected)
    return;

  ChartWidget* chart = new ChartWidget(ui->graphsListWidget, dialog->getSpec());

  charts.push_back(chart);

  ui->graphsListWidget->addItem(chart);
  ui->graphsListWidget->setItemWidget(chart, chart->getView());

  for (QString title : dialog->getList()) {
    QLineSeries* series = new QLineSeries();
    series->setName(title);
    chart->getChart()->addSeries(series);
    for (QAbstractAxis* axis : chart->getChart()->axes()) {
      series->attachAxis(axis);
    }
  }
}

void MainWindow::on_action_Load_triggered() {
  if (!fileDialog->exec())
    return;

  for (QString file : fileDialog->selectedFiles()) {
    try {
      Plugin* plugin = new Plugin(file.toStdString());
      plugin->init(&Graph::GUI::addData);
      pluginModel->add(plugin);
    } catch (std::runtime_error e) {
      std::cout << e.what() << std::endl;
      QMessageBox::critical(this, "Error", "Failed to load plugin:\n" + file, QMessageBox::StandardButton::Ignore);
    }

    int row = pluginModel->rowCount();
    pluginModel->insertRow(row);
    QModelIndex index = pluginModel->index(row);
    ui->pluginsListView->doItemsLayout();
    ui->pluginsListView->setCurrentIndex(index);
  }
}

} // namespace Graph::GUI
